#include <stdio.h>
#include <stdlib.h>


struct item {
    void* info;
    struct item *next;
};

typedef struct item Item;

struct handler {
    int size;
    Item *head, *tail;
};

typedef struct handler Handler;

/*CREA UN HANDLER E RESTITUISCE IL PUNTATORE*/
Handler* createHandler(){
    Handler *h;
    h = malloc(sizeof(Handler));
    h->size = 0;
    h->head = NULL;
    h->tail = NULL;
    return(h);
};


/*INSERISCE UN VALORE CREANDO UN OGG IN TESTA*/
void push(Handler *h, void *k) {
    Item* new;

    h->size++;
    new = malloc(sizeof(Item));
    new->info = k;
    new->next = h->head;
    h->head = new;
    if(h->tail == NULL) h->tail = new;
};

/*INSERISCE UN VALORE CREANDO UN OGG IN CODA*/
void enqueue(Handler *h, void*k){
    Item* new;

    h->size++;
    new = malloc(sizeof(Item));
    new->info = k;
    new->next = NULL;

    if(h->tail == NULL){
        h->tail = new;
        h->head = new;
    }else{
        h->tail->next = new;
        h->tail = new;
    }
};

/*ESTRAE IL VALORE DAL PRIMO OGG E LO CANCELLA*/
void* dequeue(Handler *h){
    void *v;
    Item *old;

	if(h->size>0){
		h->size--;
		v = (h->head)->info;
		old = h->head;
		h->head = old->next;
		free(old);
		return(v);
	}else printf("lista vuota\n");
};

/*Inserisce ordinatamente un nuovo valore nella lista*/
void insert(Handler *h, void* k) {
	Item *new, *previous, *current;

	new = malloc(sizeof(Item));

	h->size++;

	new->info = k;
	new->next = NULL;

	previous = NULL;
	current = h->head;

	while (current != NULL && k > current->info) {
		previous = current;              /* passa al... */
	    current = current->next;      /* ... prossimo nodo */
	}

    if (previous == NULL){
	  new->next = h->head;
	  h->head = new;
	}else{
	  previous->next = new;
	  new->next = current;
	}
};


/*STAMPA UNA LISTA*/
void printlist(Handler *h){
    Item *temp;

	if(h->size > 0){
		temp = h->head;
		while(NULL != temp){
			printf("%d ",(int)(temp->info));
			temp = temp->next;
		}
		printf("\n");
	}else printf("lista vuota\n");
};

/*INVERTE UNA LISTA*/
void invertList(Handler *h){
    Item *prec=NULL, *x=NULL, *succ=NULL;

    if((!(h->head)) || (!((h->head)->next))) return;
    x = h->head;
    succ = (h->head)->next;
    while(x != NULL){
        x->next = prec;
        prec = x;
        x = succ;
        if(x!=NULL) succ = x->next;
    }
    succ = h->head;
    h->head = h->tail;
    h->tail = succ;
};

/*SPOSTA LA SECONDA LISTA ALLA FINE DELLA PRIMA*/
void link(Handler *h1, Handler *h2){
    if (!h2) return;
	(h1->tail)->next = h2->head;
    h1->tail = h2->tail;
    free(h2);
};

void del(Handler *h, int k){
    Item *temp, *succ;
    int i;

	if((k <= h->size) && (h->size > 0)){
		h->size--;
		temp = h->head;
		if(k == 1){
			h->head = temp->next;
			free(temp);
		}else{
			for(i=2;i<k;i++) temp = temp->next;
			succ = temp->next;
			temp->next = succ->next;
			free(succ);
		}
		temp = h->head;
		while(temp->next!=NULL) temp = temp->next;
		h->tail = temp;
	}else printf("elemento non trovato oppure controllare che la lista non sia vuota\n");
};


Item* mergep(Item* L1,Item* L2){
	if(!(Item*)L1) return (Item*)L2;
	if(!(Item*)L2) return (Item*)L1;

	if ( ((Item*)L1)->info > ((Item*)L2)->info){
		((Item*)L1) -> next = mergep((Item*)L2,((Item*)L1) -> next );
		return (Item*)L1;
	}
	else{
		((Item*)L2) -> next = mergep((Item*)L1,((Item*)L2) -> next );
		return (Item*)L2;
	}
};



Item* merge_sort (Item* L){
	Item* L1=NULL;
	Item* L2=NULL;
	Item* x;
	Item* y;

	if(!(Item*)L || !((Item*)L) -> next) return (Item*)L;
	while((Item*)L){
		x = ((Item*)L) -> next;
		((Item*)L) -> next = L1;
		L1 =(Item*)L;
		if(x){
			y=x->next;
			x->next=L2;
			L2=x;
			(Item*)L=y;
		}else (Item*)L=NULL;
	}
	L1 = merge_sort (L1);
	L2 = merge_sort (L2);
	return mergep(L1,L2);
};

Handler* mergesort(Handler *h){

	h->head = merge_sort(h->head);
	return(h);

}
